/**
 * JASMINe
 * Copyright (C) 2008-2009 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbstractDroolsWorkingMemory.java 4855 2009-08-10 12:59:59Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.jasmine.rules;

import java.io.File;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.drools.KnowledgeBase;
import org.drools.builder.KnowledgeBuilder;
import org.drools.builder.KnowledgeBuilderConfiguration;
import org.drools.builder.KnowledgeBuilderFactory;
import org.drools.builder.conf.DumpDirOption;
import org.drools.runtime.StatefulKnowledgeSession;
import org.drools.runtime.rule.FactHandle;
import org.ow2.jasmine.rules.api.DroolsWorkingMemoryState;
import org.ow2.jasmine.rules.api.IDroolsWorkingMemory;

public abstract class AbstractDroolsWorkingMemory implements
        IDroolsWorkingMemory {

    /**
     * Logger.
     */
    private static Logger logger = null;

    /**
     * Thread to fire rules.
     */
    protected Thread thread = null;

    /**
     * The managed knowledge session.
     */
    protected volatile StatefulKnowledgeSession knowledgeSession = null;

    /**
     * {@link ClassLoader} to build rules.
     */
    protected ClassLoader classLoader;

    /**
     * The KnowledgeBase
     */
    protected KnowledgeBase knowledgeBase;

    /**
     * The state of the working memory
     */
    protected volatile DroolsWorkingMemoryState state;

    /**
     * The configuration used to create package builders
     */
    protected KnowledgeBuilderConfiguration knowledgeBuilderConf;


    // ------------------------------------------------------------------------
    // Constructor
    // ------------------------------------------------------------------------

    protected AbstractDroolsWorkingMemory(final Logger logger) {

        AbstractDroolsWorkingMemory.logger = logger;
        state = DroolsWorkingMemoryState.IDLE;
    }

    /**
     * @param classLoader
     */
    protected void initKnowledgeBuilderConfiguration(final ClassLoader classLoader) {

        Properties props = new Properties();
        props.setProperty("drools.dialect.java.compiler", "JANINO");
        this.knowledgeBuilderConf =
            KnowledgeBuilderFactory.newKnowledgeBuilderConfiguration(props, classLoader);

        // In development, dump drools generated files
        if(Boolean.getBoolean("jonas.developer")) {
            File dumpDir = new File(System.getProperty("java.io.tmpdir")
                    + File.separator + "drools_dump");
            boolean exists = dumpDir.exists();
            if(!exists) {
                exists = dumpDir.mkdir();
            }
            if(exists && dumpDir.isDirectory()) {
                knowledgeBuilderConf.setOption(DumpDirOption.get(dumpDir));
            }
        }
    }


    // ------------------------------------------------------------------------
    // Implementation of the IDroolsWorkingMemory interface
    // ------------------------------------------------------------------------

    /**
     * Stop the WorkingMemory, with the state STOPPED.
     * @see {@link AbstractDroolsWorkingMemory#stop(DroolsWorkingMemoryState)}
     */
    /* (non-Javadoc)
     * @see org.ow2.jasmine.rules.api.IDroolsWorkingMemory#stop()
     */
    public void stop() {
        stop(DroolsWorkingMemoryState.STOPPED);
    }

    /**
     * Stop the working memory with a given state (allowed states are
     * STOPPED and IDLE)
     */
    public void stop(DroolsWorkingMemoryState state) {
        if (knowledgeSession != null) {
            knowledgeSession.halt();
        }
        if (thread != null) {
            thread.interrupt();
            thread = null;
        }
        knowledgeSession = null;
        if( state == DroolsWorkingMemoryState.IDLE
         || state == DroolsWorkingMemoryState.STOPPED ) {
            this.state = state;
        } else {
            this.state = DroolsWorkingMemoryState.STOPPED;
        }
    }

    /**
     * Insert a fact in the WorkingMemory
     * @param fact the fact object to insert
     * @return the corresponding FactHandle
     */
    public FactHandle insert(final Object fact) {
        if (state != DroolsWorkingMemoryState.STOPPED && knowledgeSession != null) {
            return knowledgeSession.insert(fact);
        } else {
            return null;
        }
    }

    // ------------------------------------------------------------------------
    // Protected methods
    // ------------------------------------------------------------------------

    /**
     * Create a new working memory for a given rule base.
     *
     * @param ruleBase
     *            rule base
     */
    protected void createWorkingMemory() {
        logger.log(Level.FINE, "Init the knowledge session...");
        knowledgeSession = knowledgeBase.newStatefulKnowledgeSession();
    }

    /**
     * Start the rule execution process
     */
    protected void startRuleExecutor() {
        state = DroolsWorkingMemoryState.RUNNING;
        thread = new Thread(new RuleExecutor());
        thread.start();
    }

    /**
     * Periodically fire rules.
     */
    private class RuleExecutor implements Runnable {

        /**
         * Execute the rules after the timeout
         */
        public void run() {
            knowledgeSession.fireUntilHalt();
        }
    }

    /**
     * Get the KnowledgeBase
     * @return the KnowledgeBase
     */
    public KnowledgeBase getKnowledgeBase() {
        return knowledgeBase;
    }

    /**
     * Get the current state of the working memory
     * @return the state
     * @see {@link org.ow2.jasmine.rules.api.DroolsWorkingMemoryState}
     */
    public DroolsWorkingMemoryState getState() {
        return state;
    }

    /**
     * Get a PackageBuilder to compile a rule package externally
     * @return an initialized PackageBuilder
     */
    public KnowledgeBuilder newKnowledgeBuilder() {
        return KnowledgeBuilderFactory.newKnowledgeBuilder(knowledgeBuilderConf);
    }
}
